{-|
Module      : Gargantext.Core.Worker
Description : Asynchronous worker logic
Copyright   : (c) CNRS, 2024
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

-}

{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE UndecidableInstances #-}
{-# OPTIONS_GHC -Wno-orphans      #-}  -- orphan HasNodeError IOException


module Gargantext.Core.Worker where


import Async.Worker.Broker.PGMQ (PGMQBroker)
import Async.Worker.Broker.Types (BrokerMessage, toA, getMessage, messageId)
import Async.Worker qualified as W
import Async.Worker.Types qualified as W
import Async.Worker.Types (HasWorkerBroker)
import Data.Text qualified as T
import Gargantext.API.Admin.Auth (forgotUserPassword)
import Gargantext.API.Admin.Auth.Types (ForgotPasswordAsyncParams(..))
import Gargantext.API.Node.Corpus.New (addToCorpusWithForm, addToCorpusWithQuery)
import Gargantext.API.Node.New (postNode')
import Gargantext.Core.Config (hasConfig, gc_jobs)
import Gargantext.Core.Config.Types (jc_max_docs_scrapers)
import Gargantext.Core.Config.Worker (WorkerDefinition(..))
import Gargantext.Core.Worker.Broker (initBrokerWithDBCreate)
import Gargantext.Core.Worker.Env
import Gargantext.Core.Worker.Jobs.Types (Job(..), getWorkerMNodeId)
import Gargantext.Core.Worker.Types (JobInfo(..))
import Gargantext.Database.Query.Table.User (getUsersWithEmail)
import Gargantext.Prelude
import Gargantext.System.Logging ( logLocM, LogLevel(..), logMsg, withLogger )
import Gargantext.Utils.Jobs.Error (HumanFriendlyErrorText(..))
import Gargantext.Utils.Jobs.Monad (MonadJobStatus(markStarted, markComplete, markFailed))
import System.Posix.Signals (Handler(Catch), installHandler, keyboardSignal)


initWorkerState :: (HasWorkerBroker PGMQBroker Job)
                => WorkerEnv
                -> WorkerDefinition
                -> IO (W.State PGMQBroker Job)
initWorkerState env (WorkerDefinition { .. }) = do
  let gargConfig = env ^. hasConfig
  broker <- initBrokerWithDBCreate gargConfig

  pure $ W.State { broker
                 , queueName = _wdQueue
                 , name = T.unpack _wdName
                 , performAction = performAction env
                 , onMessageReceived = Just $ notifyJobStarted env
                 , onJobFinish = Just $ notifyJobFinished env
                 , onJobTimeout = Just $ notifyJobTimeout env
                 , onJobError = Just $ notifyJobFailed env
                 , onWorkerKilledSafely = Just $ notifyJobKilled env }

notifyJobStarted :: (HasWorkerBroker PGMQBroker Job)
                 => WorkerEnv
                 -> W.State PGMQBroker Job
                 -> BrokerMessage PGMQBroker (W.Job Job)
                 -> IO ()
notifyJobStarted env (W.State { name }) bm = do
  let j = toA $ getMessage bm
  let job = W.job j
  withLogger () $ \ioL ->
    logMsg ioL DEBUG $ "[notifyJobStarted] [" <> name <> "] starting job: " <> show j
  let ji = JobInfo { _ji_message_id = messageId bm
                   , _ji_mNode_id = getWorkerMNodeId job }
  let jh = WorkerJobHandle { _w_job_info = ji }
  runWorkerMonad env $ markStarted 1 jh

notifyJobFinished :: (HasWorkerBroker PGMQBroker Job)
                  => WorkerEnv
                  -> W.State PGMQBroker Job
                  -> BrokerMessage PGMQBroker (W.Job Job)
                  -> IO ()
notifyJobFinished env (W.State { name }) bm = do
  let j = toA $ getMessage bm
  let job = W.job j
  withLogger () $ \ioL ->
    logMsg ioL DEBUG $ "[notifyJobFinished] [" <> name <> "] finished job: " <> show j
  let ji = JobInfo { _ji_message_id = messageId bm
                   , _ji_mNode_id = getWorkerMNodeId job }
  let jh = WorkerJobHandle { _w_job_info = ji }
  runWorkerMonad env $ markComplete jh

notifyJobTimeout :: (HasWorkerBroker PGMQBroker Job)
                 => WorkerEnv
                 -> W.State PGMQBroker Job
                 -> BrokerMessage PGMQBroker (W.Job Job)
                 -> IO ()
notifyJobTimeout env (W.State { name }) bm = do
  let j = toA $ getMessage bm
  let job = W.job j
  withLogger () $ \ioL ->
    logMsg ioL ERROR $ "[notifyJobTimeout] [" <> name <> "] job timed out: " <> show j
  let ji = JobInfo { _ji_message_id = messageId bm
                   , _ji_mNode_id = getWorkerMNodeId job }
  let jh = WorkerJobHandle { _w_job_info = ji }
  runWorkerMonad env $ markFailed (Just $ UnsafeMkHumanFriendlyErrorText "Worker job timed out!") jh

notifyJobFailed :: (HasWorkerBroker PGMQBroker Job, HasCallStack)
                 => WorkerEnv
                 -> W.State PGMQBroker Job
                 -> BrokerMessage PGMQBroker (W.Job Job)
                 -> SomeException
                 -> IO ()
notifyJobFailed env (W.State { name }) bm exc = do
  let j = toA $ getMessage bm
  let job = W.job j
  withLogger () $ \ioL ->
    logMsg ioL ERROR $ "[notifyJobFailed] [" <> name <> "] failed job: " <> show j <> "   --- ERROR: " <> show exc
  let ji = JobInfo { _ji_message_id = messageId bm
                   , _ji_mNode_id = getWorkerMNodeId job }
  let jh = WorkerJobHandle { _w_job_info = ji }
  runWorkerMonad env $ markFailed (Just $ UnsafeMkHumanFriendlyErrorText "Worker job failed") jh

notifyJobKilled :: (HasWorkerBroker PGMQBroker Job, HasCallStack)
                 => WorkerEnv
                 -> W.State PGMQBroker Job
                 -> Maybe (BrokerMessage PGMQBroker (W.Job Job))
                 -> IO ()
notifyJobKilled _ _ Nothing = pure ()
notifyJobKilled env (W.State { name }) (Just bm) = do
  let j = toA $ getMessage bm
  let job = W.job j
  withLogger () $ \ioL ->
    logMsg ioL ERROR $ "[notifyJobKilled] [" <> name <> "] failed job: " <> show j
  let ji = JobInfo { _ji_message_id = messageId bm
                   , _ji_mNode_id = getWorkerMNodeId job }
  let jh = WorkerJobHandle { _w_job_info = ji }
  runWorkerMonad env $ markFailed (Just $ UnsafeMkHumanFriendlyErrorText $ "Worker '" <> T.pack name <> "' was killed") jh


-- | Spawn a worker with PGMQ broker
-- TODO:
--  - reduce size of DB pool
--  - progress report via notifications
--  - I think there is no point to save job result, as usually there is none (we have side-effects only)
--  - replace Servant.Job to use workers instead of garg API threads
withPGMQWorker :: (HasWorkerBroker PGMQBroker Job)
                => WorkerEnv
                -> WorkerDefinition
                -> (Async () -> W.State PGMQBroker Job -> IO ())
                -> IO ()
withPGMQWorker env wd cb = do
  state' <- initWorkerState env wd

  withAsync (W.run state') (\a -> cb a state')

withPGMQWorkerSingle :: (HasWorkerBroker PGMQBroker Job)
                     => WorkerEnv
                     -> WorkerDefinition
                     -> (Async () -> W.State PGMQBroker Job -> IO ())
                     -> IO ()
withPGMQWorkerSingle env wd cb = do
  state' <- initWorkerState env wd

  withAsync (W.runSingle state') (\a -> cb a state')


withPGMQWorkerCtrlC :: (HasWorkerBroker PGMQBroker Job)
                    => WorkerEnv
                    -> WorkerDefinition
                    -> (Async () -> W.State PGMQBroker Job -> IO ())
                    -> IO ()
withPGMQWorkerCtrlC env wd cb = do
  withPGMQWorker env wd $ \a state' -> do
    let tid = asyncThreadId a
    _ <- installHandler keyboardSignal (Catch (throwTo tid W.KillWorkerSafely)) Nothing
    cb a state'

withPGMQWorkerSingleCtrlC :: (HasWorkerBroker PGMQBroker Job)
                          => WorkerEnv
                          -> WorkerDefinition
                          -> (Async () -> W.State PGMQBroker Job -> IO ())
                          -> IO ()
withPGMQWorkerSingleCtrlC env wd cb = do
  withPGMQWorkerSingle env wd $ \a state' -> do
    let tid = asyncThreadId a
    _ <- installHandler keyboardSignal (Catch (throwTo tid W.KillWorkerSafely)) Nothing
    cb a state'


-- | How the worker should process jobs
performAction :: (HasWorkerBroker PGMQBroker Job)
              => WorkerEnv
              -> W.State PGMQBroker Job
              -> BrokerMessage PGMQBroker (W.Job Job)
              -> IO ()
performAction env _state bm = do
  let job' = toA $ getMessage bm
  let job = W.job job'
  let ji = JobInfo { _ji_message_id = messageId bm
                   , _ji_mNode_id = getWorkerMNodeId job }
  let jh = WorkerJobHandle { _w_job_info = ji }
  case job of
    Ping -> runWorkerMonad env $ do
      $(logLocM) DEBUG "[performAction] ping"
    AddCorpusFormAsync { .. } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "[performAction] add corpus form"
      addToCorpusWithForm _acf_user _acf_cid _acf_args jh
    AddCorpusWithQuery { .. } -> runWorkerMonad env $ do
      $(logLocM) DEBUG "[performAction] add corpus with query"
      let limit = Just $ fromIntegral $ env ^. hasConfig . gc_jobs . jc_max_docs_scrapers
      addToCorpusWithQuery _acq_user _acq_cid _acq_args limit jh
    ForgotPasswordAsync { _fpa_args = ForgotPasswordAsyncParams { email } } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "[performAction] forgot password: " <> email
      us <- getUsersWithEmail (T.toLower email)
      case us of
        [u] -> forgotUserPassword u
        _ -> pure ()
    NewNodeAsync { .. } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "[performAction] new node async "
      void $ postNode' _nna_authenticatedUser _nna_node_id _nna_postNode
    GargJob { _gj_garg_job } -> runWorkerMonad env $ do
      $(logLocM) DEBUG $ "Garg job: " <> show _gj_garg_job <> " (handling of this job is still not implemented!)"
      return ()
